
"""
Extended base class for fenpei job that runs through an executable (incl. shell script);
this should be considered abstract.

As with :ref:Job, your custom job(s) should inherit from this job and extend the relevant methods.
Instead of :ref: prepare and :ref: start, you can override:

* get_files
* run_file
"""

from socket import gethostname
from collections import Mapping
from os import listdir, symlink
from os.path import join, basename, isdir, isfile, dirname, exists, islink
from shutil import copyfile
from bardeen.system import mkdirp
from fenpei.job import Job
from fenpei.shell import run_shell
from datetime import datetime
from time import time

from fenpei.utils import substitute, FormattingException
from .shell import git_current_hash


def extend_substitutions(subst, name, batch, directory, git_hash=None):
	timestr = datetime.now().strftime('%Y-%m-%d %H:%M') + ' (%d)' % time()
	if git_hash is None:
		git_hash = git_current_hash()
	if isinstance(subst, Mapping):
		subst['name'] = name
		subst['batch_name'] = batch
		subst['now'] = timestr
		subst['directory'] = directory
		subst['hostname'] = gethostname()
		subst['git_commit'] = git_hash
	elif subst is None:
		pass
	else:
		raise NotImplementedError('I haven\'t thought about this, maybe it\'s not needed anyway.')


class ShJob(Job):

	def __init__(self, name, substitutions, weight=1, batch_name=None, formatter='jinja2', use_symlink=True, force_node=None):
		"""
		Create a executable or shell job object, provided a number of files or directories which will be copied,
		and (optionally) substitutions for each of them.

		:param substitutions: a dictionary; keys are files or directories to be copied, values are dicts of
		substitutions, or None; e.g. {'run.sh': {'R': 15, 'method': 'ccsd(t)'}} will copy
		:raise ShJob.FileNotFound: subclass of OSError, indicating the files argument contains data that is invalid

		For other parameters, see :ref: Job.

		Files will be copied if bool(files) is True (for substitutions), otherwise it will be attempted to
		hard-link them (use True to prevent that with no substitutions); if you use a directory, /path/ copies
		files from it and /path copies the directory with files; directory substitutions apply to contained files.
		"""
		assert ' ' not in self.run_file(), 'there should be no whitespace in run file'
		super(ShJob, self).__init__(name=name, weight=weight, batch_name=batch_name, force_node=force_node)
		git_commit = git_current_hash()
		for filepath, subst in substitutions.items():
			extend_substitutions(subst, name, batch_name, self.directory, git_hash=git_commit)
		self.files = {filepath: None for filepath in self.get_files()}
		self.files.update(substitutions)
		self.formatter = formatter
		self.use_symlink = use_symlink
		self.files = self._fix_files(self.files)

	@classmethod
	def get_files(cls):
		"""
		:return: the list of files and directories used by this code, which will be linked or copied

		Substitutions for non-static files should be supplied to the constructor.
		"""
		raise NotImplementedError()

	@classmethod
	def run_file(cls):
		"""
		:return: the path to the file which executes the job; should be in get_files()

		Files can be either string paths or tuples of (directory, pathname); in the later case pathname will be
		copied (including directories), instead of assuming only the file is to be copied.
		"""
		raise NotImplementedError()

	class FileNotFound(OSError):
		"""
		File was not found exception.
		"""

	def _fix_files(self, files):
		"""
		Check that self.files is filled with valid values (files exist etc).

		Turns all string filepaths into tuples of directory and filename.

		Expands all directories into lists of files.

		:raise ShJob.FileNotFound: subclass of OSError, indicating the one of the files doesn't exist
		"""
		def expand_dir(pre_path, source_pth, target_pth):
			"""
			Expand a tuple (predir, postdir) into all the files in that directory.
			"""
			fullpath = join(pre_path, source_pth)
			subs = []
			if isdir(fullpath):
				for subpath in listdir(fullpath):
					subs.extend(expand_dir(pre_path=pre_path,
						source_pth= join(fullpath, subpath).lstrip(pre_path),
						target_pth=target_pth))
				return subs
			else:
				return [(pre_path, source_pth, target_pth)]

		newfiles = {}
		for filepath, subst in files.items():
			if not isinstance(filepath, tuple) and not isinstance(filepath, list):
				""" change string path into tuple """
				filepath = dirname(filepath), basename(filepath), basename(filepath)
			if len(filepath) == 2:
				filepath = filepath[0], filepath[1], filepath[1]
			for pth_triple in expand_dir(*filepath):
				""" recursively expand a directory into all it's files """
				newfiles[pth_triple] = subst
			if not exists(join(*filepath[:2])):
				raise self.FileNotFound('"%s" is not a valid file or directory' % join(*filepath[:2]))
		return newfiles

	def is_prepared(self):
		"""
		See if prepared by checking the existence of every file.
		"""
		for fromroot, frompth, topth in self.files.keys():
			if not isfile(join(self.directory, topth)) and not islink(join(self.directory, topth)):
				self._log('{0:s} is not prepared because {1:s} (and possibly more) are missing'.format(self.name, frompth), 3)
				return False
		if not isfile(join(self.directory, self.run_file())):
			return False
		return True

	def prepare(self, *args, **kwargs):
		"""
		Prepares the job for execution by copying or linking all the files, and substituting values where applicable.
		"""
		super(ShJob, self).prepare(*args, **kwargs)
		if self.is_prepared():
			return False
		for (fromroot, frompth, topth), subst in self.files.items():
			sourcefilepath = join(fromroot, frompth)
			destfilepath = join(self.directory, topth)
			if exists(destfilepath):
				self._log('{0:s} is not prepared but already has file {1:s}'.format(self.name, destfilepath), 2)
				break
			mkdirp(join(self.directory, dirname(topth)))
			if subst:
				""" copy files and possibly substitute """
				if isinstance(subst, Mapping):
					with open(sourcefilepath, 'r') as fhr:
						with open(destfilepath, 'w+') as fhw:
							inp = fhr.read()
							try:
								if hasattr(self.formatter, '__call__'):
									# noinspection PyCallingNonCallable
									outp = self.formatter(inp, subst, job=self, filename=sourcefilepath)
								else:
									outp = substitute(inp, subst, formatter=self.formatter, job=self, filename=sourcefilepath)
								fhw.write(outp)
							except FormattingException as err:
								self._log('{0:}'.format(err))
								self.cleanup()
								return False
				else:
					if self.use_symlink:
						symlink(sourcefilepath, destfilepath)
					else:
						copyfile(sourcefilepath, destfilepath)
			else:
				""" Soft-link files if possible and allowed by settings """
				if self.use_symlink:
					symlink(sourcefilepath, destfilepath)
				else:
					copyfile(sourcefilepath, destfilepath)
		if isfile(join(self.directory, self.run_file())):
			run_shell(cmd = 'chmod ug+x "%s"' % join(self.directory, self.run_file()), wait = True)
		else:
			raise self.FileNotFound(('.run_file() "%s" not found after preparation; make sure it\'s origin is in ' +
				'.get_files() or in __init__ substitutions argument') % self.run_file())
		return True

	def start(self, node, *args, **kwargs):
		"""
		Start the job and store node/pid.
		"""
		self._start_pre(*args, **kwargs)
		""" nohup, bg and std redirect should be handeled by queue """
		cmd = './{0:s}'.format(self.run_file())
		pid = self.queue.run_cmd(job = self, cmd = cmd)
		self._start_post(node, pid, *args, **kwargs)
		return True






"""
	test feinpei
"""

from fenpei.queue_local import LocalQueue
from fenpei.test.job_test import TestJob


def test_jobs():

	jobs = []
	params = set(int(k**1.7) for k in range(1, 15))

	for N in params:
		jobs.append(TestJob(
			name = 'test%d' % N,
			subs = {'N': N},
			weight = int(N / 10) + 1,
		))

	queue = LocalQueue()
	queue.all_nodes()
	queue.add_jobs(jobs)
	return queue


if __name__ == '__main__':
	queue = test_jobs()
	queue.run_argv()






from copy import deepcopy
from numpy import ndarray
from numpy.random import choice, seed as set_seed


def param_sample_random(sample_space, count=100, seed=None):
	"""
	:param sample_space: A dictionary with either
		- Scalar values, which will be used as-is.
		- List, tuple, set or array, which will be chosen from at random (uniform).
		- Scipy distributions, which will be sampled from. [NOT IMPLEMENTED YET]
	:param count: How many samples
	:return: An iterable of samples.
	"""
	# possible improvement: remove duplicates if any
	if seed is not None:
		set_seed(seed)
	sampled = []
	for nr in range(count):
		sample = type(sample_space)()
		for key, value in sample_space.items():
			if isinstance(value, (int, float, str)):
				sample[key] = value
			elif isinstance(value, (list, tuple, set, ndarray)):
				sample[key] = choice(value)
			elif hasattr(value, 'rvs'):
				sample[key] = value.rvs()
			else:
				raise ValueError('param_sample_random does not know what do with object of type {0:}'
					.format(type(value)))
		sampled.append(sample)
	return sampled


def param_sample_grid(sample_space, max_count=10000):
	"""
	:param sample_space: A dictionary with either
		- Scalar values, which will be used as-is.
		- List, tuple, set or array, which will be chosen from at random (uniform).
	:param max_count: Maximum count after which to stop. Happens in chunks so may overshoot.
	:return: An iterable of all samples.
	"""
	def expand(sample, key, only_first=False):
		for val in set(sample.get(key)):
			gen = deepcopy(sample)
			gen[key] = val
			yield gen
			if only_first:
				break
	grid = [sample_space]
	for key, value in sample_space.items():
		# print('>> key', key)
		if isinstance(value, (int, float, str)):
			pass
		elif isinstance(value, (list, tuple, set, ndarray)):
			newgrid = []
			for sample in grid:
				only_first = len(newgrid) >= max_count
				newgrid.extend(expand(sample, key, only_first=only_first))
			grid = newgrid
		elif hasattr(value, 'rvs'):
			raise ValueError('param_sample_grid does not accept probability distributions, try param_sample_random'
				.format(type(value)))
		else:
			raise ValueError('param_sample_grid does not know what do with object of type {0:}'
				.format(type(value)))
	return grid


# def samples_to_code(samples):
# 	"""
# 	UNUSED. Prints the generated samples as Python dictionary definitions (assumes the keys are strings).
# 	"""
# 	for sample in samples:
# 		print('dict({0:s})'.format(
# 			', '.join('{0:s}={1:}'.format(key, '"{0:s}"'.format(value) if isinstance(value, str) else value)
# 				for key, value in sample.items())))






"""
A job that represents a series of subjobs whose results should be joined.
"""

from copy import copy
from itertools import product
from fenpei.job_sh_single import ShJobSingle


class CombiSingle(ShJobSingle):
	#todo: don't actually need the 'Sh'(shell) part, but that's how the hierarchy grew
	#todo: this is kind of a hybrid between a queue and a job, which is undesirable design
	
	def __init__(self, name, subs, ranges, child_cls, batch_name=None, child_kwargs=None,
			name_template=None, aggregation_func=None, **kwargs):
		child_kwargs = child_kwargs or {}
		self._child_cls = child_cls
		self._child_jobs = []
		weight = 0
		params = ranges.keys()
		combis = product(*ranges.values())
		for combi in combis:
			childsubs = copy(subs)
			for param, value in zip(params, combi):
				childsubs[param] = value
			if name_template is None:
				nameparts = []
				for param, value in zip(params, combi):
					nameparts.append('{1:s}{2:}'.format(name, param, value))
				childname = '_'.join([name, ''] + sorted(nameparts))
			else:
				childname = name_template.format(name=name, **childsubs)
			subjob = child_cls(name=childname, subs=childsubs, batch_name=batch_name, **child_kwargs)
			self._child_jobs.append(subjob)
			weight += subjob.weight
		sub_range = copy(subs)
		sub_range.update(ranges)
		super(CombiSingle, self).__init__(name=name, subs=sub_range, batch_name=batch_name, weight=weight, **kwargs)
		# if aggregation_func is None:
		# 	aggregation_func = self.aggregate
		self.aggregation_func = aggregation_func
	
	def __repr__(self):
		return '{0:s}*{1:d}'.format(super(CombiSingle, self).__repr__(), len(self._child_jobs))
	
	def get_default_subs(self, version=1):
		return self._child_cls.get_default_subs(version=version)
	
	def _queue_children(self):
		"""
		Connect children to a queue one-way (not add them, just connect).
		"""
		for job in self._child_jobs:
			job.queue = self.queue
	
	@classmethod
	def get_files(cls):
		return []

	@classmethod
	def get_sub_files(cls):
		return []
	
	@classmethod
	def get_nosub_files(cls):
		return []
	
	@classmethod
	def run_file(cls):
		return None
	
	def is_prepared(self):
		self._queue_children()
		for job in self._child_jobs:
			if not job.is_prepared():
				self._log('{0:s} not prepared because {1:s} is not'.format(self, job), level=3)
				return False
		self._log('{0:s} prepared because all {1:d} children are'.format(self, len(self._child_jobs)), level=3)
		return True

	def is_started(self):
		#todo: optimization by doing previously running jobs last
		self._queue_children()
		for job in self._child_jobs:
			if not job.is_started():
				self._log('{0:s} not started because {1:s} is not'.format(self, job), level=3)
				return False
		self._log('{0:s} started because all {1:d} children are'.format(self, len(self._child_jobs)), level=3)
		return True

	def is_running(self):
		#todo: optimization by doing previously running jobs last
		self._queue_children()
		for job in self._child_jobs:
			if job.is_running():
				self._log('{0:s} running because {1:s} is'.format(self, job), level=3)
				return True
		self._log('{0:s} not running because not all {1:d} children are'.format(self, len(self._child_jobs)), level=3)
		return False

	def is_complete(self):
		self._queue_children()
		for job in self._child_jobs:
			if not job.is_complete():
				self._log('{0:s} not complete because {1:s} is not'.format(self, job), level=3)
				return False
		self._log('{0:s} complete because all {1:d} children are'.format(self, len(self._child_jobs)), level=3)
		return True

	def prepare(self, verbosity=0, *args, **kwargs):
		self._queue_children()
		cnt = 0
		for job in self._child_jobs:
			cnt += job.prepare(*args, verbosity=0, **kwargs)
		return bool(cnt)
	
	def start(self, node, verbosity=0, *args, **kwargs):
		self._queue_children()
		cnt = 0
		for job in self._child_jobs:
			cnt += job.start(node, *args, verbosity=0, **kwargs)
		return bool(cnt)

	def fix(self, verbosity=0, *args, **kwargs):
		self._queue_children()
		cnt = 0
		for job in self._child_jobs:
			cnt += job.fix(*args, verbosity=0, **kwargs)
		return bool(cnt)

	def kill(self, verbosity=0, *args, **kwargs):
		self._queue_children()
		cnt = 0
		for job in self._child_jobs:
			cnt += job.kill(*args, verbosity=0, **kwargs)
		return bool(cnt)

	def cleanup(self, skip_conflicts=False, verbosity=0, *args, **kwargs):
		self._queue_children()
		cnt = 0
		for job in self._child_jobs:
			cnt += job.cleanup(skip_conflicts=skip_conflicts, *args, verbosity=0, **kwargs)
		return bool(cnt)

	def result(self, *args, **kwargs):
		assert self.queue, 'cannot get results for {0:} since it doesn\'t have a queue'.format(self)
		if not self.is_complete():
			return None
		job_results = self.queue.result(jobs=self._child_jobs, parallel=False)
		if not job_results:
			return None
		if self.aggregation_func is None:
			return self.aggregate(job_results)
		return self.aggregation_func(job_results)

	def _crash_reason_if_crashed(self, verbosity=0, *args, **kwargs):
		completed, crashed = 0, 0
		for job in self._child_jobs:
			# print(job, job.find_status())
			if not job.is_complete():
				completed += 1
			if job._crash_reason_if_crashed():
				crashed += 1
		return '+' * completed + 'X' * crashed + '.' * (len(self._child_jobs) - completed - crashed)
	
	def aggregate(self, job_results):
		"""
		Combine results from child jobs. Used if `aggregation_func` argument is not provided.
		"""
		return {res['name']: res for res in job_results.values()}






"""
Base class for fenpei job; this should be considered abstract.

Your custom job(s) should inherit from this job and extend the relevant methods, such as::

* is_prepared
* is_complete
* prepare
* start
* result
* summary
"""

from re import match
from sys import stdout
from bardeen.system import mkdirp
from time import time
from os import remove
from os.path import join, isdir
from shutil import rmtree
from .utils import CALC_DIR


class Job(object):

	CRASHED, NONE, PREPARED, RUNNING, COMPLETED = -1, 0, 1, 2, 3
	status_names = {-1: 'crashed', 0: 'nothing', 1: 'prepared', 2: 'running', 3: 'completed'}

	queue = None
	node = None
	pid = None
	status = None
	''' Set a group_cls to report results together with another class (that has the same output format). '''
	group_cls = None

	def __init__(self, name, weight=1, batch_name=None, force_node=None):
		"""
		Create a Job object.

		:param name: unique name consisting of letters, numbers, dot (.) and underscore (_) **YOU need to make sure that name is unique (bijectively maps to job)**
		:param weight: the relative resource use of this job (higher relative weights means fewer jobs will be scheduled together)
		:param batch_name: optionally, a name of the same format as ``name``, which specifies the batch (will be grouped)
		:param force_node: demand a specific node; it's up to the queue whether this is honoured
		"""
		assert match(r'^\w[/\w\.\+_-]*$', name), 'This is not a valid name: "{0:}"'.format(name)
		assert weight > 0
		self.name = name
		self.weight = weight
		self.cluster = None
		self.batch_name = batch_name
		self.force_node = force_node
		if self.batch_name:
			assert match('^\w[\w\._-]*$', batch_name)
			self.directory = join(CALC_DIR, batch_name, name)
		elif batch_name is None:
			raise AssertionError('no batch name for {0:}; provide batch_name argument when creating jobs or set it to False'.format(self))
		elif batch_name is False:
			self.directory = join(CALC_DIR, name)
		self.status = self.NONE

	def __repr__(self):
		if hasattr(self, 'name'):
			return self.name
		return '{0:s} id{1:}'.format(self.__class__.__name__, id(self))

	def _log(self, txt, *args, **kwargs):
		"""
		Logging function.
		.queue is not always set, so have own logging function.
		"""
		if self.queue is None:
			if len(txt.strip()):
				stdout.write(txt + '\n')
			else:
				stdout.write('(empty)\n')
		else:
			self.queue._log(txt, *args, **kwargs)

	def save(self):
		"""
		Save information about a running job to locate the process.
		"""
		assert self.node is not None
		assert self.pid is not None
		with open('%s/node_pid.job' % self.directory, 'w+') as fh:
			fh.write('%s\n%s\n%s\n%s' % (self.name, self.node, self.pid, str(time())))
		self._log('job %s saved' % self, level = 3)

	def unsave(self):
		"""
		Remove the stored process details.
		"""
		try:
			remove('%s/node_pid.job' % self.directory)
		except IOError:
			pass
		self._log('job %s save file removed' % self.name, level = 3)

	def load(self):
		"""
		Load process details from cache.
		"""
		try:
			with open('%s/node_pid.job' % self.directory, 'r') as fh:
				lines = fh.read().splitlines()
				self.node = lines[1]
				self.pid = int(lines[2])
			self._log('job %s loaded' % self.name, level = 3)
			return True
		except IOError:
			self._log('job %s save file not found' % self, level = 3)
			return False

	def is_prepared(self):
		pass

	def is_started(self):
		if not self.is_prepared():
			return False
		l = self.load()
		return l

	def is_running(self):
		"""
		Only called if at least prepared.
		"""
		if self.pid is None:
			if not self.load():
				return False
		if not self.queue:
			raise Exception('cannot check if %s is running because it is not in a queue' % self)
		proc_list = self.queue.processes(self.node)
		try:
			return self.pid in [proc['pid'] for proc in proc_list if proc is not None]
		except KeyError:
			raise Exception('node %s for job %s no longer found?' % (self.node, self))

	def is_complete(self):
		"""
		Check if job completed successfully.

		Needs to be extended by child class.

		Only called for jobs that are at least prepared.
		"""
		return True

	def find_status(self):
		"""
		Find status using is_* methods.
		"""
		def check_status_indicators(self):
			if self.is_prepared():
				if self.is_complete():
					return self.COMPLETED
				elif self.is_started():
					if self.is_running():
						return self.RUNNING
					return self.CRASHED
				return self.PREPARED
			return self.NONE
		if time() - getattr(self, '_last_status_time', time() - 100) > 0.7:
			self.status = check_status_indicators(self)
			setattr(self, '_last_status_time', time())
		return self.status

	def status_str(self):
		return self.status_names[self.find_status()]

	def prepare(self, silent=False, *args, **kwargs):
		"""
		Prepares the job for execution.

		More steps are likely necessary for child classes.
		"""
		self.status = self.PREPARED
		if not self.is_prepared():
			if self.batch_name:
				mkdirp(join(CALC_DIR, self.batch_name))
			mkdirp(self.directory)
		if not silent:
			self._log('preparing {0:s}'.format(self), level=2)
		""" child method add more steps here """

	def _start_pre(self, *args, **kwargs):
		"""
		Some checks at the beginning of .start().
		"""
		if self.is_running() or self.is_complete():
			if not self.queue is None:
				if self.queue.force:
					if self.is_running():
						self.kill()
				else:
					raise AssertionError('you are trying to restart a job that is running or completed; ' + \
						'use restart (-e) to skip such jobs or -f to overrule this warning')
		if not self.is_prepared():
			self.prepare(silent=True)

	def _start_post(self, node, pid, *args, **kwargs):
		"""
		Some bookkeeping at the end of .start().
		"""
		self.node = node
		self.pid = pid
		self.save()
		if self.is_running():
			self.STATUS = self.RUNNING
		self._log('starting %s on %s with pid %s' % (self, self.node, self.pid), level = 2)

	def start(self, node, *args, **kwargs):
		"""
		Start the job and store node/pid.
		"""
		self._start_pre(*args, **kwargs)
		"""
		Your starting code here.
		"""
		self._start_post(node, 'pid_here', *args, **kwargs)
		return True

	def fix(self, *args, **kwargs):
		"""
		Some code that can be ran to fix jobs, e.g. after bugfixes or updates.

		Needs to be implemented by children for the specific fix applicable (if just restarting is not viable).
		"""
		return False

	def kill(self, *args, **kwargs):
		"""
		Kills the current job if running using queue methods.

		Any overriding should probably happen in :ref: queue.processes and :ref: queue.stop_job.
		"""
		if self.is_running():
			assert self.node is not None
			assert self.pid is not None
			self._log('killing %s: %s on %s' % (self, self.pid, self.node), level = 2)
			self.queue.stop_job(node = self.node, pid = self.pid)
			return True
		self._log('job %s not running' % self, level = 3)
		return False

	def cleanup(self, skip_conflicts=False, *args, **kwargs):
		if self.is_running() or self.is_complete():
			if self.queue is not None and not self.queue.force:
				if skip_conflicts:
					return False
				raise AssertionError('you are trying to clean up a job that is running or completed; ' + \
					'use -f to force this, or -e to skip these jobs (it could also mean that two ' + \
					'jobs are use the same name and batchname).')
		if self.batch_name is not False and isdir(self.directory):
			rmtree(self.directory, ignore_errors = True)
			self._log('cleaned up {0:s}'.format(self), level=2)
			return True
		return False

	def result(self, *args, **kwargs):
		"""
		Collects the result of the completed job.

		:return: result of the job; only requirement is that the result be compatible with :ref: summary (and other jobs), but a dict is suggested.
		"""
		if not self.is_complete():
			return None
		return None

	def crash_reason(self, verbosity=0, *args, **kwargs):
		"""
		Find the reason the job has crashed. Should only be called for crashed jobs (by _crash_reason_if_crashed).
		"""
		if verbosity <= 0:
			return '??'
		else:
			return '?? reason for crash not known'

	def _crash_reason_if_crashed(self, verbosity=0, *args, **kwargs):
		if not self.find_status() == self.CRASHED:
			return None
		return self.crash_reason(verbosity=verbosity, *args, **kwargs)






import sys
from os import getcwd
from subprocess import Popen, PIPE

"""
	Keep a list of processes to stop them from being terminated if their reference goes out of scope.
"""
process_memory = []


# only tested with wait = True
def run_shell(cmd, wait):
	if wait:
		process = Popen(cmd, shell=True, stdout=PIPE, stderr=PIPE)
		outp, err = process.communicate()
		if err:
			sys.stderr.write(err.strip())
			return None
		return outp
	else:
		devnull = open('/dev/null', 'w')
		process = Popen(cmd, shell=True, stdout=devnull, stderr=sys.stderr)
		process_memory.append(process)
	return None


def run_cmds_on(cmds, node, wait=True, queue=None):
	"""
	Run several commands on a local or remote machine.

	:param cmds: list of command lines, non Popen style words
	:param node:
	:param wait:
	:param queue:
	:return: list of stdout for each command if succesful, None otherwise
	:raise: no exceptions; writes to sys.stderr for problems

	commands need to be merged because otherwise the state of cd is forgotten
	"""
	split_str = '#%&split_here&%#'
	cmds = [cmd.strip() for cmd in cmds]
	cmds = [cmd[:-1] if cmd.endswith(';') else cmd for cmd in cmds]
	cmd_str = ('; echo \'%s\'; ' % split_str).join(cmds)
	cmd_str = cmd_str.replace('\\"', '"').replace('"', '\\"').replace('&; ', '& ')
	if node is None:
		cmd_str = ('bash -c "%s"' % cmd_str).replace('\n', '')
	else:
		cmd_str = ('ssh %s "%s"' % (node, cmd_str)).replace('\n', '')
	if queue:
		queue._log(cmd_str.replace('echo \'%s\'; ' % split_str, ''), level=3)
	raw_outp = run_shell(cmd_str, wait=wait)
	if raw_outp is None:
		return None
	outp = [block.strip() for block in raw_outp.split(split_str)]
	return outp


def run_cmds(cmds, wait=True, queue=None):
	return run_cmds_on(cmds, node=None, wait=wait, queue=queue)


def git_current_hash():
	def getit():
		process = Popen('git rev-parse --verify HEAD', shell=True, stdout=PIPE, stderr=PIPE)
		outp, err = process.communicate()
		if err:
			return '[no git commit found]'
		return outp.strip()
	setattr(git_current_hash, '_CACHE', getattr(git_current_hash, '_CACHE', {}))
	if getcwd() not in git_current_hash._CACHE:
		git_current_hash._CACHE[getcwd()] = getit()
	return git_current_hash._CACHE[getcwd()]


# if __name__ == '__main__':
# 	print(git_current_hash())






"""
Queue using qsub to start jobs.
"""
from distutils.spawn import find_executable
from logging import warning
from os import popen, environ
from os.path import join
from repoze.lru import lru_cache
from fenpei.shell import run_cmds
from fenpei.queue import Queue
from re import findall


class SlurmQueue(Queue):

	def __init__(self, jobs=None, partition=None, summary_func=None):
		self.partition = partition or 'thchem'
		super(SlurmQueue, self).__init__(jobs=jobs, summary_func=summary_func)
		self.time_limit = '07-00:00:00'
		if find_executable('sinfo'):
			partition_info = run_cmds(['sinfo -l --partition {0:s}'.format(self.partition)], queue=self)
			if partition_info:
				partition_lines = partition_info[0].splitlines()[2:]
				if partition_lines:
					time = partition_lines[0].split()[2]
					if findall(r'^[-\d]*\d+:\d+:\d+$', time):
						self.time_limit = time

	def all_nodes(self):
		"""
		Specific nodes are irrelevant; everything in main queue.
		"""
		if not super(SlurmQueue, self).all_nodes():
			return False
		self._log('no specific nodes; all to general queue')
		self.nodes = [self.partition]
		return True

	def node_availability(self):
		raise NotImplementedError('this should not be implemented for %s because the qsub-queue does the distributing' % self.__class__)

	def distribute_jobs(self, jobs = None, max_reject_spree = None):
		"""
		Let slurm do the distributing by placing everything in general queue.
		"""
		self._log('call to distribute for %d jobs ignored; slurm will do distribution' % len(jobs))
		self.all_nodes()
		self.distribution = {0: jobs}

	@lru_cache(10)
	def test_slurm(self):
		if not find_executable('sinfo'):
			self._log('slurm does not work on this machine; run this code from a node that has access to the queue')
			exit()

	def get_slurm_stat(self):
		"""
		Get slurm status for current user as a dictionary of properties.
		"""
		with popen('squeue --partition {0:s} --user $USER --format \'%A %B %P %T %u %j\''.format(self.partition)) as fh:
			txt = fh.read()
		parts = [line.split(None, 5) for line in txt.splitlines()[1:]]
		jobs = []
		for taskinfo in parts:
			if taskinfo[3] in ('PENDING', 'RUNNING', 'SUSPENDED', 'COMPLETING', 'COMPLETED'):
				jobs.append({
					'pid': int(taskinfo[0]),
					'name': taskinfo[5],
					'user': taskinfo[4],
					'queue': taskinfo[2],
					'node': taskinfo[1],
					'state': taskinfo[3],
				})
			else:
				warning(('task with id {0:s} is still in slurm queue, but has state {1:s} ' +
					'(it will be assumed to have crashed)').format(taskinfo[0], taskinfo[3]))
		return jobs

	@lru_cache(maxsize=100, timeout=2.5)
	def processes(self, node):
		"""
		Get process info from qstat.
		"""
		self.test_slurm()
		self._log('loading processes for %s' % node, level = 3)
		return self.get_slurm_stat()

	def stop_job(self, node, pid):
		"""
		Remove individual job from queue.
		"""
		run_cmds(['scancel {0:d}'.format(pid)], queue=self)

	def run_cmd(self, job, cmd):
		"""
		Start an individual job by means of queueing a shell command.
		"""
		self.test_slurm()
		assert job.directory
		node_flags = ()
		comment = 'batch: {0:s}; job: {1:s}; weight: {2:d}'.format(job.batch_name, job.name, job.weight)
		if job.force_node:
			node_flags += (
				'--nodelist', str(job.force_node),
				'--no-requeue',
			)
			comment = '{0:s}; forced to node: {1:s}'.format(comment, job.force_node)
		elif 'EXCLUDE_NODES' in environ and environ['EXCLUDE_NODES'].strip():
			node_flags += (
				'--exclude', '"{0:}"'.format(environ['EXCLUDE_NODES']),
			)
			comment = '{0:s}; excl: {1:s}'.format(comment, environ['EXCLUDE_NODES'].strip())
		core_flags = (
			'sbatch',
			'--job-name', '"{0:s}"'.format(job.name),
			'--partition', str(self.partition),
			'--workdir', '"{0:s}"'.format(job.directory),
			'--time', self.time_limit,
			'--mem', '{0:d}G'.format(2+job.weight),
			'--ntasks', '1',  # different tasks can be on different nodes
			'--cpus-per-task', str(max(min(job.weight//1, 12), 1)),
			'--nodes', '1',
			'--output', '"{0:s}"'.format(join(job.directory, 'slurm.all')),
			'--error',  '"{0:s}"'.format(join(job.directory, 'slurm.all')),
		)
		if getattr(job, 'niceness', True):
			node_flags += (
				'--nice={0:}'.format(getattr(job, 'niceness', 100)),  # otherwise other people can't run
			)
			comment = '{0:s}; nice: {1:}'.format(comment, getattr(job, 'niceness', 100))
		core_flags += ('--comment', '"{0:s}"'.format(comment))
		subcmd = ' '.join(core_flags + node_flags + ('\'{0:s}\''.format(cmd),))
		cdcmd = 'cd "{0:s}"'.format(job.directory)
		outp = run_cmds((cdcmd, subcmd,), queue=self)
		self._log(subcmd, level=3)
		if not outp or not outp[1]:
			raise self.CmdException('job {0:s} could not be queued (output is empty)'.format(job))
		qid = findall(r'Submitted batch job (\d+)(\s|$)', outp[1])[0][0]
		if not qid:
			raise self.CmdException('job {0:s} id could not be found in "{1:s}"'.format(job, outp[1]))
		return int(qid)






"""
Extended base class for fenpei job that runs through a Python file; this should be considered abstract.

As with :ref:Job, your custom job(s) should inherit from this job and extend the relevant methods.
Instead of :ref: prepare and :ref: start, you can override:

* run_template : template python file to copy
* run_file : location of the copied python file to be executed

It has become clear to me that this is easier when wrapped in a shell script with ShJobSingle instead.
"""

from os.path import join, exists, split
from bardeen.system import mkdirp, link_else_copy
from fenpei.job import Job
from .utils import CALC_DIR


class PyJob(Job):

	def is_prepared(self):
		return exists(self.run_file())

	def run_template(self):
		"""
		Return the path to the Python file to run (which is then copied and ran or added to a queue or something).
		"""
		return self.template_file

	def run_file(self):
		"""
		Return the path to the link/copy of .run_template(), whether or not it exists.
		"""
		return join(self.directory, split(self.run_template())[1])

	def prepare(self, *args, **kwargs):
		"""
		Prepares the job for execution.

		Creates directory and copies run_template to run_file.

		More steps are likely necessary for child classes.
		"""
		self.status = self.PREPARED
		if not self.is_prepared():
			mkdirp(self.directory)
		link_else_copy(self.run_template(), self.run_file())

	def start(self, node, *args, **kwargs):
		"""
		Start the job and store node/pid.
		"""
		self._start_pre(*args, **kwargs)
		pid = self.queue.run_job(node = node, filepath = self.run_file())
		self._start_post(node, pid, *args, **kwargs)
		return True






"""
Like ShJob, but with one set of substitutions for all files (or None).

Automatically adds all substitutions as attributes to the job.
"""

from base64 import urlsafe_b64encode
from collections import OrderedDict
from copy import copy
from json import dump, load
from logging import warning
from os import remove
from os.path import join, exists
from struct import pack
from fenpei.job_sh import ShJob, extend_substitutions
from fenpei.shell import run_cmds


class ShJobSingle(ShJob):

	def __init__(self, name, subs, sub_files=(), nosub_files=(), weight=1, batch_name=None,
			defaults_version=1, formatter='jinja2', skip_checks=False, use_symlink=True, force_node=None):
		"""
		Similar to ShJob.

		:param subs: a dictionary of substitutions (not specific to files, contrary to ShJob).
		:param files: files to which substitutions should be applied.
		:param nosub_files: files as-is (no substitutions).
		:param defaults_version: which version of defaults? (Exists to keep old jobs working).
		:param formatter: which formatter to use for files (%, .format, jinja, ..; see `utils.py`).
		"""
		""" Defaults for substitutions. """
		subs_with_defaults = copy(self.get_default_subs(version=defaults_version))
		self.parameter_names = list(subs_with_defaults.keys())
		for sub in subs.keys():
			if sub not in subs_with_defaults:
				warning('job "{0:}" has unknown substitution parameter "{1:s}" = "{2:}"'.format(self, sub, subs[sub]))
		subs_with_defaults.update(subs)
		subs_with_defaults['defaults_version'] = defaults_version
		""" Check/make sure that combiantions of parameters are acceptable """
		checked_subs = self.check_and_update_subs(subs_with_defaults, skip_checks=skip_checks)
		""" Substitutions as job properties. """
		self.substitutions = checked_subs
		for key, val in checked_subs.items():
			setattr(self, key, val)
		""" Override the whole ShJob init because it's very inefficient if all substitutions are the same """
		""" This skips one inheritance level! """
		super(ShJob, self).__init__(name=name, weight=weight, batch_name=batch_name, force_node=force_node)
		self.formatter = formatter
		self.use_symlink = use_symlink
		extend_substitutions(self.substitutions, name, batch_name, self.directory)
		if not hasattr(self.__class__, '_FIXED_CACHE'):
			""" Create the (path, name) -> subst map, but use True instead of the map. """
			files = {filepath: None for filepath in self.get_nosub_files() + list(nosub_files)}
			files.update({filepath: True for filepath in self.get_sub_files() + list(sub_files)})
			self.__class__._FIXED_CACHE = self._fix_files(files)
		""" Now fill in the substitutions (in a copied version). """
		self.files = {fileinfo: (copy(self.substitutions) if subs is True else None) for (fileinfo, subs) in self.__class__._FIXED_CACHE.items()}
		self.parameter_file_path = join(self.directory, 'parameters.json')

	def _calc_param_hash(self):
		param_data = hash(tuple(self.substitutions[nm] for nm in self.parameter_names))
		return urlsafe_b64encode(pack('!q', param_data)).rstrip('=')

	@property
	def param_hash(self):
		ShJobSingle._HASH_CACHE = getattr(ShJobSingle, '_HASH_CACHE', {})
		if id(self) not in ShJobSingle._HASH_CACHE:
			ShJobSingle._HASH_CACHE[id(self)] = self._calc_param_hash()
		return ShJobSingle._HASH_CACHE[id(self)]

	def check_and_update_subs(self, subs, *args, **kwargs):
		return subs

	@classmethod
	def get_default_subs(cls, version = 1):
		"""
		:return: default values for substitutions
		"""
		return OrderedDict()

	@classmethod
	def get_files(cls):
		"""
		(used by ShJob; make sure jobs are not added twice)
		"""
		return []

	@classmethod
	def get_sub_files(cls):
		"""
		:return: list of files with substitutions
		"""
		return []

	@classmethod
	def get_nosub_files(cls):
		"""
		:return: list of files without substitutions
		"""
		return []

	def get_input(self):
		subfiles = self.get_sub_files()
		if subfiles:
			return self.files[subfiles[0]]
		return None

	def prepare(self, verbosity=0, *args, **kwargs):
		status = super(ShJobSingle, self).prepare(verbosity=verbosity, *args, **kwargs)
		self.store_config()
		chcmd = 'chmod 750 -R "{0:s}"'.format(self.directory)
		outp = run_cmds((chcmd,), queue=self.queue)
		return status

	def fix(self, verbosity=0, force=False, *args, **kwargs):
		is_fixed = super(ShJobSingle, self).fix(verbosity=verbosity, *args, **kwargs)
		if self.is_prepared() and not exists(self.parameter_file_path):
			self.store_config()
			self._log('stored parameters for {0:}'.format(self), level=2)
			is_fixed = True
		try:
			self.check_config()
		except AssertionError as err:
			if force:
				remove(self.parameter_file_path)
				self.store_config()
				self._log('overwriting stored parameters for {0:} (because of -f)'.format(self), level=2)
				is_fixed = True
			else:
				warning(str(err) + '; use -f to replace the file')
		return is_fixed

	def result(self, verbosity=0, *args, **kwargs):
		self.check_config()
		return super(ShJobSingle, self).result(verbosity=verbosity, *args, **kwargs)

	def crash_reason(self, verbosity=0, *args, **kwargs):
		self.check_config()
		return super(ShJobSingle, self).crash_reason(verbosity=verbosity, *args, **kwargs)

	def store_config(self):
		store = OrderedDict()
		for name in self.parameter_names:
			store[name] = self.substitutions[name]
		with open(self.parameter_file_path, 'w+') as fh:
			dump(obj=store, fp=fh, indent=0)

	def check_config(self):
		if not self.is_prepared():
			return
		try:
			with open(self.parameter_file_path, 'r') as fh:
				retrieved = load(fp=fh)
		except (OSError, IOError) as err:
			raise AssertionError(('Job {0:} has no loadable stored parameters for consistency checking; ' +
				'they can be created using -g [load error: {1:}]').format(self, err))
		before, now = set(retrieved.keys()), set(self.parameter_names)
		assert not (before - now), 'settings disappeared compared to when job was run: {0:} (-gf to reset this check)'\
			.format(', '.join(before - now))
		assert not (now - before), 'new settings appeared compared to when job was run: {0:} (-gf to reset this check)'\
			.format(', '.join(now - before))
		for name in self.parameter_names:
			if not retrieved[name] == self.substitutions[name]:
				raise AssertionError('parameter {0:} for job {1:} was initially an {2:} <{3:}> but is now {4:} <{5:}>'
					.format(name, self, type(retrieved[name]).__name__, retrieved[name],
					type(self.substitutions[name]).__name__, self.substitutions[name]))






"""
	Queue using qsub to start jobs.
"""

from os import popen
from os.path import join
from repoze.lru import lru_cache
from fenpei.shell import run_cmds
from fenpei.queue import Queue
from re import findall
from xml.dom.minidom import parse


class QsubQueue(Queue):

	DEFAULT_QSUB_NAME = 'queuename'

	def __init__(self, jobs=None, qname=None, summary_func=None):
		self.qname = qname or self.DEFAULT_QSUB_NAME
		super(QsubQueue, self).__init__(jobs=jobs, summary_func=summary_func)

	def all_nodes(self):
		"""
		Specific nodes are irrelevant; everything in main queue.
		"""
		self._log('no specific nodes; all to general queue', level=2)
		if self.load_nodes():
			return False
		self.nodes = [self.qname]
		self.slots = []
		return True

	def node_availability(self):
		raise NotImplementedError('this should not be implemented for %s because the qsub-queue does the distributing' % self.__class__)

	def distribute_jobs(self, jobs = None, max_reject_spree = None):
		"""
		Let qsub do the distributing by placing everything in general queue.
		"""
		self._log('call to distribute for %d jobs ignored; qsub will do distribution' % len(jobs), level=2)
		self.all_nodes()
		self.distribution = {0: jobs}

	@lru_cache(10)
	def _test_qstat(self):
		if run_cmds(['qstat'], queue = self) is None:
			self._log('qstat does not work on this machine; run this code from a node that has access to the queue')
			exit()

	def _get_qstat(self):
		"""
		Get qstat for current user as a dictionary of properties.

		Based on http://stackoverflow.com/questions/26104116/qstat-and-long-job-names
		"""
		f = popen('qstat -xml -r')
		dom = parse(f)
		jobelem = dom.getElementsByTagName('job_info')
		joblist = jobelem[0].getElementsByTagName('job_list')
		jobs = []
		for job in joblist:
			jobstate = job.getElementsByTagName('state')[0].childNodes[0].data
			try:
				qstatqueue = job.getElementsByTagName('queue_name')[0].childNodes[0].data
			except IndexError:
				qstatqueue = '(no queue yet)'
			try:
				node = qstatqueue.split('@')[1].split('.')[0]
			except IndexError:
				node = None
			if 'E' in jobstate:
				jobs.append(None)
			jobs.append({
				'pid': int(job.getElementsByTagName('JB_job_number')[0].childNodes[0].data),
				'name': job.getElementsByTagName('JB_name')[0].childNodes[0].data,
				'user': job.getElementsByTagName('JB_owner')[0].childNodes[0].data,
				'queue': qstatqueue,
				'node': node,
				'state': jobstate,
			})
		return jobs

	@lru_cache(maxsize=100, timeout=2.5)
	def processes(self, node):
		"""
		Get process info from qstat.
		"""
		self._test_qstat()
		self._log('loading processes for %s' % node, level = 3)
		return self._get_qstat()

	def stop_job(self, node, pid):
		"""
		Remove individual job from queue.
		"""
		run_cmds(['qdel %s' % pid], queue = self)

	def run_cmd(self, job, cmd):
		"""
		Start an individual job by means of queueing a shell command.
		"""
		self._test_qstat()
		queue = self.qname
		if job.force_node:
			queue = '{0:s}@{1:s}'.format(self.qname, job.force_node)
			self._log('job {0:s} forced queue {1:s}'.format(job, queue), level=2)
		assert job.directory
		subcmd = [
			'qsub',                             # wait in line
				'-b', 'y',                      # it's a binary
				'-cwd',                         # use the current working directory
				'-q', queue,                    # which que to wait in
				'-N', job.name,                 # name of the job
				#'-l slots={0:d}'.format(job.weight), # number of slots = weight of job
					#check this; maybe it's threads rather than processes
				'-e', join(job.directory, 'qsub.err'),  # error directory for the queue
				'-o', join(job.directory, 'qsub.out'),  # output directory for the queue
			'bash -c \'%s\'' % cmd,		    # the actual command (single quotes!)
		]
		cmds = [
			'cd \'%s\'' % job.directory,
			' '.join(subcmd),
		]
		outp = run_cmds(cmds, queue = self)
		self._log(cmds[-1], level = 3)
		if not outp or not outp[1]:
			raise self.CmdException('job %s could not be started (output is empty)' % job)
		qid = findall(r'Your job (\d+) \("[^"]+"\) has been submitted', outp[1])[0]
		if not qid:
			raise self.CmdException('job %s id could not be found in "%s"' % (job, outp[1]))
		return int(qid)






from functools import partial
from multiprocessing import Pool, cpu_count
from tempfile import gettempdir
#from threading import Thread
from warnings import warn
from bardeen.system import mkdirp
from os import environ, chmod
from os.path import join, expanduser

from jinja2 import StrictUndefined

if 'CALC_DIR' in environ:
	CALC_DIR = environ['CALC_DIR']
else:
	CALC_DIR = join(expanduser('~'), 'data/sheffield')

TMP_DIR = join(gettempdir(), 'fenpei')
mkdirp(TMP_DIR)
chmod(TMP_DIR, 0o700)


def get_pool_light():
	"""
	Process pool for light work, like IO. (This object cannot be serialized so can't be part of Queue). Also consider thread_map.
	"""
	if not hasattr(get_pool_light, 'pool'):
		setattr(get_pool_light, 'pool', Pool(min(3 * cpu_count(), 20)))
	return getattr(get_pool_light, 'pool')


def thread_map(func, data):
	"""
	http://code.activestate.com/recipes/577360-a-multithreaded-concurrent-version-of-map/
	"""
	#todo: this is disabled
	return map(func, data)

	N = len(data)
	result = [None] * N

	# wrapper to dispose the result in the right slot
	def task_wrapper(i):
		result[i] = func(data[i])

	threads = [Thread(target=task_wrapper, args=(i,)) for i in xrange(N)]
	for t in threads:
		t.start()
	for t in threads:
		t.join()

	return result


def _make_inst(params, JobCls, default_batch=None):
	if 'batch_name' not in params and default_batch:
		params['batch_name'] = default_batch
	return JobCls(**params)


def create_jobs(JobCls, generator, parallel=True, default_batch=None):
	"""
	Create jobs from parameters in parallel.
	"""
	if parallel:
		jobs = get_pool_light().map(partial(_make_inst, JobCls=JobCls, default_batch=default_batch), tuple(generator))
	else:
		jobs = list(_make_inst(params, JobCls=JobCls, default_batch=default_batch) for params in generator)
	return jobs


def substitute(text, substitutions, formatter, job=None, filename=None):
	"""
	If `formatter` is not callable, this routine is called to choose a built-in formatter.
	"""
	if formatter in [None, 'none']:
		return text
	elif formatter in ['%', 'pypercent']:
		return substitute_pypercent(text, substitutions, job=job, filename=filename)
	elif formatter in ['.format', 'pyformat']:
		return substitute_pyformat(text, substitutions, job=job, filename=filename)
	elif formatter == 'jinja2':
		return substitute_jinja2(text, substitutions, job=job, filename=filename)
	elif formatter == 'jinja':
		warn('Using jinja2 for formatter `jinja`; change to `jinja2` to suppress this warning')
		return substitute_jinja2(text, substitutions, job=job, filename=filename)
	else:
		raise NotImplemented('formatter "{0:}" not known; use a callable or one of `pypercent`, `pyformat` or `jinja2`'.format(formatter))


class FormattingException(Exception): pass


def substitute_pypercent(text, substitutions, job=None, filename=None):
	"""
	Use old Python % formatting to apply substitutions to a string.
	"""
	outp = []
	for nr, line in enumerate(text.splitlines()):
		try:
			outp.append((line % substitutions))
		except KeyError as err:
			raise FormattingException('missing key "{0:s}" in substitution of "{1:s}" on line {2:d}; job not prepared'
				.format(str(err).strip('\''), filename, nr + 1))
		except ValueError:
			raise FormattingException('substitution of "%s" on line %d encountered a formatting error; job not prepared'
				.format(filename, nr + 1))
	return '\n'.join(outp)


def substitute_pyformat(text, substitutions, job=None, filename=None):
	"""
	Use new Python .format() to apply substitutions to a string.
	"""
	outp = []
	for nr, line in enumerate(text.splitlines()):
		try:
			outp.append(line.format(**substitutions))
		except KeyError as err:
			raise FormattingException('missing key "{0:s}" in substitution of "{1:s}" on line {2:d}; job not prepared'
				.format(str(err).strip('\''), filename, nr + 1))
		except ValueError:
			raise FormattingException('substitution of "%s" on line %d encountered a formatting error; job not prepared'
				.format(filename, nr + 1))
	return '\n'.join(outp)


def substitute_jinja2(text, substitutions, job=None, filename=None):
	"""
	Use `jinja2` so apply formatting to a string.
	
	:param text: A string, like the contents of a file, in which substitutions should be applied.
	:param substitutions: Also called 'context', contains a mapping of things to replace.
	:return: Substituted string.
	"""
	try:
		from jinja2 import Template, __version__ as jinja_version
	except ImportError as err:
		raise ImportError('Jinja2 is set as the formatter, but '.format(err))
	if int(jinja_version.split('.')[1]) < 7:
		raise ImportError('Jinja2 needs at least version 2.7, but you have {0:s}'.format(jinja_version))
	from jinja2 import TemplateSyntaxError
	
	try:
		template = Template(text, undefined=StrictUndefined, trim_blocks=True, lstrip_blocks=True)
	except TemplateSyntaxError as err:
		raise TemplateSyntaxError('In file {0:s}: {1:}'.format(filename, err), err.lineno)
	return template.render(**substitutions)






"""
Distribute jobs over multiple machines by means of ssh.

- find quiet nodes
- start jobs if space
- weight jobs
- get status info
- restart failed
"""

from fnmatch import fnmatch
from logging import warning
from subprocess import PIPE
from subprocess import Popen
from sys import stdout, stderr
from time import time, sleep
from random import sample
from bardeen.inout import reprint
from datetime import datetime
from collections import defaultdict, OrderedDict
from argparse import ArgumentParser, SUPPRESS
from os import remove
from os.path import basename, join
from math import ceil
from functools import partial
from .shell import run_cmds_on, run_cmds
from .job import Job
from .utils import get_pool_light, TMP_DIR, thread_map


def job_task_run(job, method, **kwargs):
	"""
	Runs an arbitrary method of job; used by job_task.
	"""
	return getattr(job, method)(**kwargs)


def job_task(method, **kwargs):
	"""
	Returns a function that runs an arbitrary method of an object, for passing to Pool.map
	"""
	return partial(job_task_run, method=method, **kwargs)


class Queue(object):

	def __init__(self, jobs=None, summary_func=None):
		self.show = 1
		self.force = False
		self.restart = False
		# self.all = False
		self.weight = None
		self.limit = None
		self.jobs = []
		self.nodes = []
		self.slots = []
		self.distribution = {}
		self.process_list = {}
		self.process_time = {}
		self.parallel = None
		if jobs:
			self.add_jobs(jobs)
		if summary_func is not None:
			self.summary = summary_func

	def _log(self, txt, level=1):
		"""
		Report to user.
		"""
		if level <= self.show:
			stdout.write(txt + '\n')

	def all_nodes(self):
		"""
		Get a list of all nodes (their ssh addresses).
		"""
		if self.load_nodes():
			return False
		self._log('finding nodes')
		self.nodes = []
		self.slots = []
		''' find node ssh adresses and store in self.nodes '''
		return True

	def node_availability(self):
		"""
		Check the processor use of all nodes.
		"""
		if self.load_nodes():
			return False
		if not len(self.nodes):
			self._log('no nodes yet; calling all_nodes()', level = 2)
			self.all_nodes()
			if not len(self.nodes):
				self._log('no nodes found; no availability checked', level = 2)
				return
		self.slots = []
		self._log('checking node availability', level = 1)
		for node in self.nodes:
			outps = run_cmds_on(cmds = ['grep \'model name\' /proc/cpuinfo | wc -l', 'uptime'], node = node, queue = self)
			if len(outps) == 2:
				''' one slot for every 100% processor available '''
				proc_count = int(outps[0])
				load_1min = float(outps[1].split()[-3].replace(',', ''))
				self.slots.append(max(proc_count - load_1min, 0))
				self._log('%2d slots assigned to %6s - 1min cpu %4d%% on %d processors' % (round(self.slots[-1]), self.short_node_name(node), 100 * load_1min, proc_count), level = 2)
			else:
				''' not accessible for some reason '''
				self._log('%s not accessible' % node)
				self.nodes.remove(node)
		self._log('found %d idle processors on %d nodes' % (sum(self.slots), len(self.nodes)))
		self.save_nodes()
		return True

	def save_nodes(self):
		"""
		Save the list of nodes to cache.
		"""
		with open('%s/timestamp.nodes' % TMP_DIR, 'w+') as fh:
			fh.write(str(time()))
		with open('%s/names.nodes' % TMP_DIR, 'w+') as fh:
			fh.write('\n'.join(self.nodes))
		with open('%s/slots.nodes' % TMP_DIR, 'w+') as fh:
			fh.write('\n'.join(['%.4f' % slot for slot in self.slots]))
		self._log('nodes saved')

	def unsave_nodes(self):
		"""
		Remove cached node data.
		"""
		try:
			remove('%s/timestamp.nodes' % TMP_DIR)
			remove('%s/names.nodes' % TMP_DIR)
			remove('%s/slots.nodes' % TMP_DIR)
			self._log('removing stored node info')
		except OSError:
			pass

	def load_nodes(self, memory_time = 10 * 60):
		"""
		Load use restart (-e) to skip such jobs othe list of nodes from cache, if not expired.
		"""
		try:
			with open('%s/timestamp.nodes' % TMP_DIR, 'r') as fh:
				timestamp = float(fh.read())
				dt = time() - timestamp
		except IOError:
			self._log('no stored node info found', level=2)
			return False
		if dt < memory_time:
			self._log('loaded node info (age: %ds)' % dt)
		else:
			self._log('stored node info outdated (%ds)' % dt)
			return False
		with open('%s/names.nodes' % TMP_DIR, 'r') as fh:
			self.nodes = fh.read().split()
		with open('%s/slots.nodes' % TMP_DIR, 'r') as fh:
			self.slots = [float(slot) for slot in fh.read().split()]
		return True

	def distribute_jobs(self, jobs = None, max_reject_spree = None):
		"""
		Distribute jobs favourably by means of kind-of-Monte-Carlo (only favourable moves).

		:param jobs: (optional) the jobs to be distributed; uses self.jobs if not provided
		:param max_reject_spree: (optional) stopping criterion; stop when this many unfavourable moves tried in a row
		:return: distribution, a dictionary with node *indixes* as keys and lists of jobs on that node as values
		"""
		if not len(self.slots) > 0:
			self.node_availability()
		if jobs is None:
			jobs = self.jobs
		assert len(self.nodes) == len(self.slots)
		max_reject_spree = 2 * len(self.nodes) if max_reject_spree is None else max_reject_spree
		self._log('distributing %d jobs with weight %d over %d slots' % (len(jobs), self.total_weight(jobs), sum(self.slots)))
		def cost(weight_1, slots_1, weight_2, slots_2):
			return max(weight_1 - slots_1, 0) ** 2 + max(weight_2 - slots_2, 0) ** 2 + slots_1 / max(weight_1, 1) + slots_2 / max(weight_2, 1)
		''' clear the list '''
		distribution = {}
		for node_nr in range(len(self.nodes)):
			distribution[node_nr] = []
		''' random initial job distribution '''
		for job in jobs:
			node_nr = sample(distribution.keys(), 1)[0]
			distribution[node_nr].append(job)
		''' repeat switching until nothing favourable is found anymore '''
		reject_spree, steps = 0, 0
		while reject_spree < 100:
			node1, node2 = sample(distribution.keys(), 2)
			if len(distribution[node1]) > 0:
				steps += 1
				cost_before = cost(self.total_weight(distribution[node1]), self.slots[node1],
								   self.total_weight(distribution[node2]), self.slots[node2])
				item1 = sample(range(len(distribution[node1])), 1)[0]
				cost_switch = cost_move = None
				if len(distribution[node2]) > 0:
					''' compare the cost of switching two items '''
					item2 = sample(range(len(distribution[node2])), 1)[0]
					cost_switch = cost(self.total_weight(distribution[node1]) - distribution[node1][item1].weight + distribution[node2][item2].weight, self.slots[node1],
									   self.total_weight(distribution[node2]) + distribution[node1][item1].weight - distribution[node2][item2].weight, self.slots[node2])
				if cost_before > 0:
					''' compare the cost of moving an item '''
					cost_move = cost(self.total_weight(distribution[node1]) - distribution[node1][item1].weight, self.slots[node1],
									 self.total_weight(distribution[node2]) + distribution[node1][item1].weight, self.slots[node2])
				''' note that None < X for any X, so this works even if only cost_before has an actual value '''
				if (cost_switch < cost_before and cost_switch is not None) or (cost_move < cost_before and cost_move is not None):
					if cost_switch < cost_move and cost_switch is not None:
						''' switch '''
						tmp = distribution[node1][item1]
						distribution[node1][item1] = distribution[node2][item2]
						distribution[node2][item2] = tmp
					elif cost_move is not None:
						''' move (move if equal, it's easier after all) '''
						distribution[node2].append(distribution[node1][item1])
						del distribution[node1][item1]
					reject_spree = 0
				else:
					''' not favorable; don't move '''
					reject_spree += 1
			else:
				''' too many empty slots means few rejectsbut lots of iterations, so in itself a sign to stop '''
				reject_spree += 0.1
		self.distribution = distribution
		''' report results '''
		self._log('distribution found after %d steps' % steps)
		self._log(self.text_distribution(distribution), level = 2)

	def text_distribution(self, distribution):
		"""
		Text visualisation of the distribution of jobs over nodes.
		"""
		lines = []
		no_job_nodes = []
		line_len_guess = max(max(self.total_weight(node_jobs) for node_jobs in distribution.values()), self.slots[0]) + 8
		for node_nr, jobs in distribution.items():
			if len(jobs):
				prog_ind, steps = '', 0
				for job in jobs:
					for k in range(int(round(job.weight - 1))):
						steps += 1
						if steps < self.slots[node_nr]:
							prog_ind += '+'
						else:
							prog_ind += '1'
					steps += 1
					if steps < self.slots[node_nr]:
						prog_ind += 'x'
					else:
						prog_ind += '!'
				prog_ind += '_' * int(round(self.slots[node_nr] - steps))
				prog_ind += ' ' * int(max(line_len_guess - len(prog_ind), 0))
				job_names = ', '.join(str(job) for job in jobs)
				prog_ind += job_names if len(job_names) <= 30 else job_names[:27] + '...'
				lines.append('%5s: %s' % (self.short_node_name(self.nodes[node_nr]), prog_ind))
			else:
				no_job_nodes.append(self.short_node_name(self.nodes[node_nr]))
		if len(no_job_nodes):
			lines.append('no jobs on %d nodes: %s' % (len(no_job_nodes), ', '.join(no_job_nodes)))
		return '\n'.join(lines)

	def short_node_name(self, long_name):
		return long_name

	def total_weight(self, jobs = None):
		"""
		Total weight of the provided jobs, or the added ones if None.
		"""
		if jobs is None:
			jobs = self.jobs
		return sum([job.weight for job in jobs])

	def processes(self, node):
		"""
		Get processes on specific node and cache them.
		"""
		if node in self.process_time.keys():
			if time() - self.process_time[node] < 3:
				return self.process_list[node]
		self._log('loading processes for %s' % node, level = 3)
		self.process_time[node] = time()
		self.process_list[node] = []
		outp = run_cmds_on([
			'ps ux',
		], node = node, queue = self)
		if outp is None:
			self._log('can not connect to %s; are you on the cluster?' % node)
			exit()
		for line in outp[0].splitlines()[1:]:
			cells = line.split()
			ps_dict = {
				'pid':  int(cells[1]),
				'name': ' '.join(cells[10:]),
				'user': cells[0],
				'start':cells[8],
				'time': cells[9],
				'node': node,
			}
			if not ps_dict['name'] == '-bash' and not ps_dict['name'].startswith('sshd: ') and not ps_dict['name'] == 'ps ux':
				self.process_list[node].append(ps_dict)
		return self.process_list[node]

	def add_job(self, job):
		"""
		Add single job to the queue.
		"""
		assert isinstance(job, Job)
		job.queue = self
		self.jobs.append(job)
		return self

	def add_jobs(self, jobs):
		"""
		Add list of jobs to the queue.
		"""
		for job in jobs:
			self.add_job(job)
		self._same_path_check()
		return self

	def get_jobs(self):
		return self.jobs

	def list_jobs(self, cols=2, verbosity=0, *args, **kwargs):
		N = int(ceil(len(self.jobs) / float(cols)))
		for k in range(N):
			stdout.write('  | '.join(
				'{0:2d}. {1:20s} {2:>10s}'.format(p + 1, '{0:s} [{1:d}]'.format(
					join(self.jobs[p].batch_name, self.jobs[p].name) if verbosity else self.jobs[p].name,
					self.jobs[p].weight
				), self.jobs[p].status_str())
					for p in [k, k+N, k+2*N] if p < len(self.jobs)
			) + '\n')

	def run_job(self, job, filepath):
		"""
		Start an individual job, specified by a Python file.
		"""
		cmd = 'nohup python \'%s\' &> out.log &' % basename(filepath)
		return self.run_cmd(job, cmd)

	class CmdException(Exception):
		"""
		An external (e.g. Popen shell script) could not be run.
		"""

	def run_cmd(self, job, cmd):
		"""
		Start an individual job by means of a shell command.

		:param job: the job that's being started this way
		:param cmd: shell commands to run (should include nohup and & as appropriate)
		:return: process id (str)
		"""
		assert job.directory
		cmds = [
			'cd \'%s\'' % job.directory,
			cmd,
			'echo "$\!"' # pid
		]
		outp = run_cmds_on(cmds, node = job.node, queue = self)
		if not outp:
			raise self.CmdException('job %s could not be started' % self)
		return str(int(outp[-1]))

	def stop_job(self, node, pid):
		"""
		Kill an individual job, specified by pid given during start ('pid' could also e.g. be a queue number).
		"""
		run_cmds_on(['kill %s' % pid], node = node, queue = self)

	def prepare(self, parallel=None, *args, **kwargs):
		"""
		Prepare all the currently added jobs (make files etc).
		"""
		parallel = self.parallel if parallel is None else parallel
		if parallel:
			#statuses = get_pool_light().map(job_task('prepare', **kwargs), self.jobs)
			statuses = thread_map(job_task('prepare', **kwargs), self.jobs)
		else:
			statuses = (job.prepare(**kwargs) for job in self.jobs)
		prepare_count = sum(int(status) for status in statuses)
		self._log('prepared %d jobs' % prepare_count)

	def running_count(self):
		"""
		How many running jobs.
		"""
		return len(self.get_status()[Job.RUNNING])

	def running_weight(self):
		"""
		Total weight of running jobs.
		"""
		return sum(job.weight for job in self.get_status()[Job.RUNNING])

	def start(self, parallel=None, verbosity=0, *args, **kwargs):
		"""
		Calls corresponding functions depending on flags (e.g. -z, -w, -q, -e).
		"""
		self._quota_warning()
		self._same_path_check(fail=True)
		# if self.all:
		# 	if self.weight:
		# 		self._log('starting all jobs; specific weight (-w) ignored')
		# 	if self.limit:
		# 		self._log('starting all jobs; limit weight (-q) ignored')
		W = float('inf')
		if self.limit:
			W = max(self.limit - self.running_weight(), 0)
			if verbosity:
				if not self.weight:
					self._log('starting jobs with weight %d (no minimum)' % W)
				elif W < self.weight:
					self._log('starting jobs with weight %d because of minimum weight %d' % (W, self.weight))
				else:
					self._log('starting jobs with weight %d to fill to %d (higher than minimum)' % (W, self.limit))
		elif self.weight:
			W = self.weight
			self._log('starting jobs with weight %d (by fixed weight)' % self.weight)
		self.start_weight(W, parallel=parallel)

	def start_weight(self, weight, parallel=None, *args, **kwargs):
		"""
		(Re)start jobs with an approximation of total weight.
		"""
		parallel = self.parallel if parallel is None else parallel
		parallel = False  # override because it's slow in this case, somehow...
		batch = []
		jobs = self.get_jobs_by_weight(weight)
		if len(jobs):
			self.distribute_jobs(jobs = jobs)
			for node_nr, jobs in self.distribution.items():
				for job in jobs:
					batch.append((job, self.nodes[node_nr]))
			if parallel:
				statuses = get_pool_light().map(partial(_start_job_on_node, **kwargs), batch)
			else:
				statuses = map(partial(_start_job_on_node, **kwargs), batch)
			start_count = sum(int(status) for status in statuses)
			self._log('(re)started %d jobs' % start_count if self.restart else 'started %d jobs' % start_count)
		else:
			self._log('no jobs to restart' if self.restart else 'no jobs to start')

	def _quota_warning(self):
		try:
			out, err = Popen(['quota', '-Q'], stdout=PIPE, stderr=PIPE).communicate()
		except OSError:
			""" No quota """
		else:
			parts = out.splitlines()[-1].split()
			used, avail = float(parts[0].strip('*')), float(parts[1].strip('*'))
			if avail - used < 1e7 and used / avail > 0.8:
				warning('there is only {0:d}MB free ({1:d}%)'.format(
					int((avail - used) / 1024), int(100 * (1 - float(used) / avail))))

	def _same_path_check(self, fail=False):
		pths = set()
		found = 0
		for job in self.jobs:
			if (job.batch_name, job.name) in pths:
				msg = 'there are multiple jobs with location {0:s}/{1:s}'.format(job.batch_name, job.name)
				if fail:
					raise AssertionError(msg)
				found += 1
				if found >= 3:
					warning(msg + '. Stopping duplicate checks now.')
					break
				warning(msg)
			pths.add((job.batch_name, job.name))
		return pths

	def get_jobs_by_weight(self, max_weight):
		"""
		Find jobs with an approximation of total weight.
		"""
		""" find eligible jobs (in specific order) """
		job_status = self.get_status()
		if self.restart:
			startable = job_status[Job.PREPARED] + job_status[Job.NONE] + job_status[Job.CRASHED]
		else:
			startable = job_status[Job.PREPARED] + job_status[Job.NONE]
		if not startable:
			self._log('there are no jobs that can be started')
			return []
		total_weight = sum(job.weight for job in startable)
		if not total_weight:
			""" start only one job """
			jobs = [startable[0]]
		elif max_weight > total_weight or max_weight is None:
			""" start all jobs """
			jobs = startable
		else:
			""" start jobs with specific weight """
			jobs, current_weight = [], 0
			startable = sorted(startable, key = lambda item: - item.weight - (10 if item.status == Job.CRASHED else 0))
			for job in startable:
				if current_weight + job.weight <= max_weight:
					jobs.append(job)
					current_weight += job.weight
		self._log('starting: ' + ', '.join(job.name for job in jobs), level = 2)
		return jobs

	def fix(self, parallel=None, *args, **kwargs):
		"""
		Fix jobs, e.g. after fixes and updates.
		"""
		parallel = self.parallel if parallel is None else parallel
		if parallel:
			statuses = thread_map(job_task('fix', **kwargs), self.jobs)
		else:
			statuses = (job.fix(**kwargs) for job in self.jobs)
		fix_count = sum(int(status) for status in statuses)
		self._log('fixed %d jobs' % fix_count)

	def kill(self, *args, **kwargs):
		"""
		Kill all the currently added job processes.
		"""
		kill_count = 0
		for job in self.jobs:
			kill_count += int(job.kill(*args, **kwargs))
		self._log('killed %d jobs' % kill_count)

	def cleanup(self, parallel=None, *args, **kwargs):
		"""
		Clean up all the currently added jobs (remove files).
		"""
		parallel = self.parallel if parallel is None else parallel
		if parallel:
			#statuses = get_pool_light().map(job_task('cleanup', **kwargs), self.jobs)
			statuses = thread_map(job_task('cleanup', skip_conflicts=self.restart, **kwargs), self.jobs)
		else:
			statuses = (job.cleanup(skip_conflicts=self.restart, **kwargs) for job in self.jobs)
		cleanup_count = sum(int(status) for status in statuses)
		self._log('cleaned up %d jobs' % cleanup_count)

	def get_status(self, parallel=None, **kwargs):
		"""
		Get list of statusses.
		"""
		parallel = self.parallel if parallel is None else parallel
		if parallel:
			status_res = thread_map(job_task('find_status', **kwargs), self.jobs)
			statuses = dict(zip(self.jobs, status_res))
		else:
			statuses = {}
			for job in self.jobs:
				statuses[job] = job.find_status(**kwargs)
		# status_count = defaultdict(int)
		status_list = defaultdict(list)
		for job, status in statuses.items():
			# status_count[status] += 1
			status_list[status].append(job)
		return status_list

	def show_status(self, status_list, verbosity=0):
		"""
		Show list of statusses.
		"""
		self._log('status for %d jobs:' % len(self.jobs), level = 1)
		for status_nr in status_list.keys():
			if verbosity <= 0:
				job_names = ' '.join(str(job) for job in status_list[status_nr][:20 ])
				job_names = job_names if len(job_names) <= 40 else job_names[:37] + '...'
				self._log(' {0:3d} {1:12s} {2:s}'.format(
					len(status_list[status_nr]), Job.status_names[status_nr], job_names))
			else:
				job_names = ' '.join(str(job) for job in status_list[status_nr])
				weight = sum(job.weight for job in self.get_status()[status_nr])
				stat_wght = '{1:s} ({0:d})'.format(weight, Job.status_names[status_nr])
				self._log(' {0:3d} {2:18s} {1:s}'.format(
					len(status_list[status_nr]), job_names, stat_wght))

	def continuous_status(self, delay=5, *args, **kwargs):
		"""
		Keep refreshing status until ctrl+C.
		"""
		self._log('monitoring status; use cltr+C to terminate')
		lines = len(Job.status_names) + 1
		stdout.write('\n' * lines + '\n')
		nothing_running_count = 0
		while True:
			try:
				status_list = self.get_status()

				txt = '%s - job# %d; weight %d:' % (datetime.now().strftime('%H:%M:%S'), self.running_count(), self.running_weight())
				for status_nr in status_list.keys():
					job_names = ', '.join(str(job) for job in status_list[status_nr])
					txt += '\n %3d %-12s %s' % (len(status_list[status_nr]), Job.status_names[status_nr], job_names if len(job_names) <= 40 else job_names[:37] + '...')
				reprint(txt, lines)

				if not status_list[Job.RUNNING]:
					nothing_running_count += 1
				else:
					nothing_running_count = 0

				if nothing_running_count >= 3:
					self._log('status monitoring terminated; no more running jobs')
					break

				""" sleep to the next %delay point (e.g. for 5s, check at :05, :10, :15 etc (not :14, :19 etc) """
				sleep(delay - (datetime.now().second + datetime.now().microsecond / 1e6 + .01) % delay)

			except KeyboardInterrupt:
				self._log('status monitoring terminated by user')
				break

	def status(self, verbosity=0, *args, **kwargs):
		"""
		Get and show the status of jobs.
		"""
		status_list = self.get_status()
		self.show_status(status_list, verbosity=verbosity)

	def result(self, parallel=None, jobs=None, *args, **kwargs):
		"""
		:return: a dict of job results, with jobs as keys.
		"""
		parallel = self.parallel if parallel is None else parallel
		# parallel = False  # override because parallel is much slower for some reason.
		if jobs is None:
			jobs = self.jobs
		results = OrderedDict()
		if parallel:
			resli = get_pool_light().map(job_task('result', **kwargs), jobs)
			for job, res in zip(jobs, resli):
				results[job] = res
		else:
			for job in jobs:
				results[job] = job.result(*args, **kwargs)
		for job, res in results.items():
			if res and 'in' not in res:
				res['in'] = job.get_input()
		self._log('retrieved results for %d jobs' % len(jobs))
		return results

	@staticmethod
	def summary(queue):
		raise NotImplementedError(('No summary function (queue "{0:s}"). Attach a '
			'static method .summary(queue) to the queue.').format(queue.name))

	def compare_jobs(self, parameters, filter=None):
		"""
		Get a parameters -> job mapping. The parameters are expected to identify unique jobs.

		:param filter: a function that returns True for jobs that should be included.
		:return: Without parameters, a list of jobs. With parameters, a mapping from parameter to accompanying jobs. Indices are parameter values for a single parameter, otherwise tuples.
		"""
		if not hasattr(parameters, '__iter__'):
			parameters = (parameters,)
		assert len(parameters) > 0, 'Provide a job attribute to compare jobs.'
		def get_key(j):
			vals = []
			for param in parameters:
				assert hasattr(job, param), 'Can not compare jobs on "{0:s}" since job "{1:s}" does not have this attribute.'.format(param, job)
				vals.append(getattr(job, param))
			# if len(vals) == 1:
			# 	return vals[0]
			return tuple(vals)
		jobmap = OrderedDict()
		if filter is None:
			filter = lambda obj: True
		for job in self.jobs:
			if filter(job):
				key = get_key(job)
				assert key not in jobmap, 'Can not compare jobs on "{0:}" since jobs "{1:s}" and "{2:s}" both have value <{3:}>, but values should be unique.'.format(parameters, jobmap[key], job, key)
				jobmap[key] = job
		return jobmap

	def compare_results(self, parameters, filter=None):
		"""
		Similar to compare_jobs but uses a map from parameters -> results instead. Furthermore, jobs without results are omitted.
		"""
		""" param -> job """
		jobmap = self.compare_jobs(parameters, filter=filter)
		""" job -> result """
		results = self.result(jobs=jobmap.values())
		""" param -> result [if complete] """
		return OrderedDict((parval, results[job]) for parval, job in jobmap.items() if results[job] is not None)

	def get_crash_reason(self, parallel=None, verbosity=0):
		"""
		For each failed job, print why it failed.
		"""
		parallel = self.parallel if parallel is None else parallel
		reasons = OrderedDict()
		if parallel:
			#resli = get_pool_light().map(job_task('_crash_reason_if_crashed', verbosity=verbosity), self.jobs)
			#resli = thread_map(job_task('_crash_reason_if_crashed', verbosity=verbosity), self.jobs)
			resli = [job._crash_reason_if_crashed(verbosity=verbosity) for job in self.jobs]
		else:
			#resli = thread_map(job_task('_crash_reason_if_crashed', verbosity=verbosity), self.jobs)
			resli = [job._crash_reason_if_crashed(verbosity=verbosity) for job in self.jobs]
		for job, res in zip(self.jobs, resli):
			if res is not None:
				reasons[job] = res
		self._log('retrieved crash reasons for %d jobs' % len(reasons))
		return reasons

	def crash_reason(self, parallel=None, verbosity=0, line_len=70, *args, **kwargs):
		reasons = self.get_crash_reason(parallel=parallel, verbosity=verbosity)
		max_name_len = max((len(job.batch_name) + len(job.name)) for job in reasons.keys()) + 1
		log_template = '{0:' + str(max_name_len) + 's} {1:s}\n'
		for job, reason in reasons.items():
			txt = [reason[k:k+line_len] for k in range(0, len(reason), line_len)]
			for k, line in enumerate(txt):
				jobname = '{0:}/{1:}'.format(job.batch_name, job.name) if k==0 else ''
				stdout.write(log_template.format(jobname, line))

	def run_cmd_per_job(self, cmd, parallel=None, verbosity=0, status=None, *args, **kwargs):
		"""
		Run a shell command in each job's directory (for --cmd).
		"""
		if status is None:
			jobs = self.jobs
		else:
			jobs = self.get_status()[status]
		cmd_count = 0
		for job in jobs:
			if job.is_prepared():
				self._log('running command for {0:s}'.format(job.name), level=2)
				scmd = 'cd \'{dir:s}\'; bash -c \'export NAME=\'"\'"\'{name:s}\'"\'"\' BATCH=\'"\'"\'{batch:s}\'"\'"\' STATUS=\'"\'"\'{status:s}\'"\'"\'; {cmd:s}\''.format(
					dir=job.directory, name=job.name, batch=job.batch_name,
					status=job.status_names[job.status], cmd=cmd
				)
				self._log(scmd, level=3)
				process = Popen([scmd], shell=True)
				process.communicate()
				cmd_count += 1
		self._log('ran command `{1:s}` for {0:d} jobs'.format(cmd_count, cmd), level=1)

	def run_argv(self):
		"""
		Analyze sys.argv and run commands based on it.
		"""
		def summary(queue = self, *args, **kwargs): return self.summary(queue)
		def wrap_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd)
		def wrap_prep_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.PREPARED)
		def wrap_run_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.RUNNING)
		def wrap_comp_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.COMPLETED)
		def wrap_crash_cmd(cmd): return partial(self.run_cmd_per_job, cmd=cmd, status=Job.CRASHED)
		parser=ArgumentParser(description='distribute jobs over available nodes',
			epilog='actions are executed (largely) in the order they are supplied; some actions may call others where necessary')
		parser.add_argument('-v', '--verbose', dest='verbosity', action='count', default=0, help='more information (can be used multiple times, -vv)')
		parser.add_argument('-f', '--force', dest='force', action='store_true', help='force certain mistake-sensitive steps instead of failing with a warning')
		parser.add_argument('-e', '--restart', dest='restart', action='store_true', help='with this, start and cleanup ignore complete (/running) jobs')
		parser.add_argument('-a', '--availability', dest='availability', action='store_true', help='list all available nodes and their load (cache reload)')
		parser.add_argument('-d', '--distribute', dest='distribute', action='store_true', help='distribute the jobs over available nodes')
		parser.add_argument('-l', '--list', dest='actions', action='append_const', const=self.list_jobs, help='show a list of added jobs')
		parser.add_argument('-p', '--prepare', dest='actions', action='append_const', const=self.prepare, help='prepare all the jobs')
		parser.add_argument('-c', '--calc', dest='actions', action='append_const', const=self.start, help='start calculating one jobs, or see -z/-w/-q')
		#parser.add_argument('-b', '--keepcalc', dest='actions', action='append_const', const=None, help='like -c, but keeps checking and filling')
		# parser.add_argument('-z', '--all', dest='all', action='store_true', help='-c will start all jobs')
		parser.add_argument('-w', '--weight', dest='weight', action='store', type=int, default=None, help='-c will start jobs with total WEIGHT running')
		parser.add_argument('-q', '--limit', dest='limit', action='store', type=int, default=None, help='-c will add jobs until a total LIMIT running')
		parser.add_argument('-k', '--kill', dest='actions', action='append_const', const=self.kill, help='terminate the calculation of all the running jobs')
		parser.add_argument('-r', '--remove', dest='actions', action='append_const', const=self.cleanup, help='clean up all the job files')
		parser.add_argument('-g', '--fix', dest='actions', action='append_const', const=self.fix, help='fix jobs, check cache etc (e.g. after update)')
		parser.add_argument('-s', '--status', dest='actions', action='append_const', const=self.status, help='show job status')
		parser.add_argument('-m', '--monitor', dest='actions', action='append_const', const=self.continuous_status, help='show job status every few seconds')
		parser.add_argument('-x', '--result', dest='actions', action='append_const', const=summary, help='run analysis code to summarize results')
		parser.add_argument('-t', '--whyfail', dest='actions', action='append_const', const=self.crash_reason, help ='print a list of failed jobs with the reason why they failed')
		parser.add_argument('-j', '--serial', dest='parallel', action='store_false', help='make job commands (start, fix, etc) serial (parallel is faster but order is inconsistent)')
		parser.add_argument('--jobs', dest='jobs', action='store', type=str, help='specify by name the jobs to (re)start or use, separated by whitespace, ?*[] patterns allowed')
		# parser.add_argument('--cmd', dest='cmd', nargs=1, action='store', type=str, help='run shell command in each of the job directories')
		parser.add_argument('--cmd', dest='actions', nargs=1, action='append', type=wrap_cmd, help='run a shell command in the directories of each job that has a dir ' + \
			'($NAME/$BATCH/$STATUS if -s); also prep-cmd, run-cmd, comp-cmd, crash-cmd')
		parser.add_argument('--prep-cmd', dest='actions', nargs=1, action='append', type=wrap_prep_cmd, help=SUPPRESS)
		parser.add_argument('--run-cmd', dest='actions', nargs=1, action='append', type=wrap_run_cmd, help=SUPPRESS)
		parser.add_argument('--comp-cmd', dest='actions', nargs=1, action='append', type=wrap_comp_cmd, help=SUPPRESS)
		parser.add_argument('--crash-cmd', dest='actions', nargs=1, action='append', type=wrap_crash_cmd, help=SUPPRESS)
		# remaining letters: bjnu  [-i, -y and -o are available but have commmon meanings]
		""" Note that some other options may be in use by subclass queues. """
		args = parser.parse_args()

		""" Handle only specific jobs by deleting the others (not the most clean perhaps, but good enough for the occasional use). """
		if args.jobs:
			requested = dict((ptrn, 0) for ptrn in args.jobs.split())
			for job in tuple(self.jobs):
				keep = False
				for ptrn in requested.keys():
					if fnmatch(job.name, ptrn):
						requested[ptrn] += 1
						keep = True
				if not keep:
					self.jobs.remove(job)
			unmatched = tuple(ptrn for ptrn, cnt in requested.items() if not cnt)
			if unmatched:
				parser.error('Specifically requested job(s) [{0:s}] was/were not found.'
					.format(', '.join(unmatched)))

		if not args.actions:
			stderr.write('No action selected. Use one or several flags to control actions.\n\n')
			parser.print_help()
			exit()

		actions = []
		for act in args.actions:
			if hasattr(act, '__iter__'):
				actions.extend(act)
			else:
				actions.append(act)

		self.show = args.verbosity + 1
		self.force, self.restart, self.weight, self.limit = \
			args.force, args.restart, args.weight, args.limit

		if not actions and not any((args.availability, args.distribute, self.restart, self.weight, self.limit,)):
			parser.print_help()
			parser.error('Please provide some action.')

		if args.availability:
			prev_show, self.show = self.show, 2
			self.unsave_nodes()
			self.all_nodes()
			self.node_availability()
			self.show = prev_show
		if args.distribute:
			self.distribute_jobs()

		if self.restart:
			if self.start not in actions and self.cleanup not in actions:
				parser.error('You requested that restart/cleanup apply only to failed jobs, '
					'but didn\'t specify a start or cleanup command [-c].')

		if self.start not in actions:
			if self.weight:
				parser.error('You specified a weight for jobs to be started, '
					'but didn\'t specify a start command [-c].')
			if self.limit:
				parser.error('You specified a weight for jobs to keep running, '
					'but didn\'t specify a start command [-c].')

		self.parallel = args.parallel
		if actions:
			for action in actions:
				action(verbosity=args.verbosity, parallel=args.parallel, force=args.force)

		return [str(action) for action in actions]


def _start_job_on_node(info, **kwargs):
	job, node = info
	# job.cleanup(**kwargs)
	return job.start(node, **kwargs)






"""
Run jobs on local machine, e.g. for testing.
"""

from fenpei.queue import Queue
from fenpei.shell import run_cmds


class LocalQueue(Queue):

	def all_nodes(self):
		self.nodes = ['localhost']
		self._log('nodes: localhost', level = 2)
		return True

	def node_availability(self):
		if not self.nodes:
			self.all_nodes()
		self.slots = [10]
		self._log('availability: localhost', level = 2)
		return True

	def distribute_jobs(self, jobs = None, max_reject_spree = None):
		if not self.slots:
			self.node_availability()
		if jobs is None:
			jobs = self.jobs
		self.distribution = {
			0: jobs
		}
		self._log('distribution: all on localhost', level = 2)

	def processes(self, node):
		"""
		Get processes on specific node and cache them.
		"""
		self._log('loading processes for %s' % node, level = 3)
		self.process_list[node] = []
		outp = run_cmds([
			'ps ux',
		], queue = self)
		for line in outp[0].splitlines()[1:]:
			cells = line.split()
			self.process_list[node].append({
				'pid':  int(cells[1]),
				'node': node,
			})
		return self.process_list[node]

	def run_cmd(self, job, cmd):
		"""
		See Queue.run_cmd(), but run everything on local machine.
		"""
		assert job.directory
		cmds = [
			'cd \'%s\'' % job.directory,
			cmd,
			'echo "$\!"' # pid
		]
		outp = run_cmds(cmds, queue = self)
		if not outp:
			raise self.CmdException('job %s could not be started' % self)
		return str(int(outp[-1]))

	def stop_job(self, node, pid):
		"""
		Kill an individual job, specified by pid given during start ('pid' could also e.g. be a queue number).
		"""
		run_cmds(['kill %s' % pid], queue = self)





# -*- coding: utf-8 -*-

from setuptools import setup


with open('README.rst', 'r') as fh:
	readme = fh.read()

setup(
	name='fenpei',
	description='Distribution of tasks.',
	long_description=readme,
	url='https://github.com/mverleg/fenpei',
	author='Mark V',
	maintainer='(the author)',
	author_email='mdilligaf@gmail.com',
	license='Revised BSD License (LICENSE.txt)',
	keywords=[],
	version='2.3',
	packages=['fenpei'],
	include_package_data=True,
	zip_safe=False,
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Developers',
		'Natural Language :: English',
		'License :: OSI Approved :: BSD License',
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Topic :: Software Development :: Libraries :: Python Modules',
	],
	install_requires=[
		'bardeen',
		'jinja2',
	],
)
